# ::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Solution proposal to programming exercises week 7       #
#                                                         #
# INF2310, spring 2017                                    #
#                                                         #
# Ole-Johan Skrede, 18.02.2017                            #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::: #

"""
Implementation of 2D convolution and separable convolutions. Application on gradient images.
"""

# pylint: disable=bad-indentation
# pylint: disable=import-error

import os

import numpy as np
import cv2
import matplotlib.pyplot as plt
from scipy import signal

def conv2d(im, kernel):
  """
  A brute force implementation of 2D convolution with zero padding.

  Args:
      im: 2D numpy array
      kernel: 2D numpy array, odd side lengths
  Returns:
      filtered_image: 2D numpy array of float64, equal size to image
  """
  M, N = im.shape
  S = int((kernel.shape[0] - 1)/2)
  T = int((kernel.shape[1] - 1)/2)
  pad_im = np.zeros((im.shape[0] + 2*S, im.shape[1] + 2*T))
  pad_im[S:-S, T:-T] = im
  filtered_image = np.zeros(im.shape, dtype=float)
  for i in range(M):
    for j in range(N):
      for s in range(2*S + 1):
        for t in range(2*T + 1):
          filtered_image[i, j] += pad_im[i - s + 2*S, j - t + 2*T]*kernel[s, t]
  return filtered_image

def conv1d_H(im, kernel):
  """Convolve rows of input im with kernel.

  Args:
      im: 2D numpy array
      kernel: 1D numpy array of length L (assumed odd)
  Returns:
      conv_im: 2D numpy array of same shape as im.
  """
  L = len(kernel)
  P = int((L - 1)/2)
  M, N = im.shape

  conv_im = np.zeros(im.shape)

  # Pad rows in image with zeros
  pad_im = np.zeros((im.shape[0], im.shape[1] + 2*P))
  pad_im[:, P:-P] = im

  # Perform convolution
  for i in range(M):
    for j in range(N):
      # Pixel (i, j) in the original image is at (i, j+P) in the padded image
      orig_vals = pad_im[i, j:j+L]
      conv_kernel = kernel[::-1] # Flip the kernel
      conv_im[i, j] = np.dot(orig_vals, conv_kernel)
  return conv_im

def conv1d_V(im, kernel):
  """Convolve columns of input im with kernel.

  Args:
      im: 2D numpy array
      kernel: 1D numpy array of length L (assumed odd)
  Returns:
      conv_im: 2D numpy array of same shape as im.
  """
  L = len(kernel)
  P = int((L - 1)/2)
  M, N = im.shape

  conv_im = np.zeros(im.shape)

  # Pad columns in image with zeros
  pad_im = np.zeros((im.shape[0] + 2*P, im.shape[1]))
  pad_im[P:-P, :] = im

  # Perform convolution
  for i in range(M):
    for j in range(N):
      # Pixel (i, j) in the original image is at (i+P, j) in the padded image
      orig_vals = pad_im[i:i+L, j]
      conv_kernel = kernel[::-1] # Flip the kernel
      conv_im[i, j] = np.dot(orig_vals, conv_kernel)
  return conv_im

def plot_image(im, fig_num, name='Image'):
  """Plot image convenience function"""

  plt.figure(fig_num)
  plt.imshow(im, cmap='gray', interpolation='none')
  plt.title(name)
  plt.axis('off')

  fig_num += 1
  return fig_num

def test_image_similarity(im1, im2, im3):
  """Check if the input images are approximately equal"""

  assert np.max(np.abs(im1 - im3)) < 1e-10, (
      "Non-similar images: {}".format(np.max(np.abs(im1 - im3))))
  assert np.max(np.abs(im2 - im3)) < 1e-10, (
      "Non-similar images: {}".format(np.max(np.abs(im2 - im3))))

def scale_uint8(im, min_val, max_val):
  """Scale image to range [min_val, max_val] and cast to uint8"""

  max_part = max_val*(im - np.min(im)) / (np.max(im) - np.min(im))
  min_part = min_val*(np.max(im) - im) / (np.max(im) - np.min(im))

  return np.round(max_part + min_part).astype(np.uint8)

def main():
  """Test implementation of convolution, and compute gradient images."""

  # Change these to suit your environment and to change images
  in_im_file = 'car.png'
  input_dir = '../../assets/images'
  result_dir = 'results/python'

  # Read image and plot it
  im = cv2.imread(os.path.join(input_dir, in_im_file), cv2.IMREAD_GRAYSCALE)
  fig_num = plot_image(im, 0, 'Original')

  # -----------------------------------
  # Vertical gradient direction (Sobel)
  # -----------------------------------
  hx = np.array([[1, 2, 1],
                 [0, 0, 0],
                 [-1, -2, -1]])
  hx_V = np.array([1, 0, -1])
  hx_H = np.array([1, 2, 1])

  gx_2d = conv2d(im, hx)
  gx_temp = conv1d_H(im, hx_H)
  gx_1d = conv1d_V(gx_temp, hx_V)
  gx_scipy = signal.convolve2d(im, hx, mode='same', boundary='fill')
  test_image_similarity(gx_1d, gx_2d, gx_scipy)

  cv2.imwrite(os.path.join(result_dir, 'grad_x_'+in_im_file), scale_uint8(gx_1d, 0, 255))

  fig_num = plot_image(gx_1d, fig_num, 'Vertical gradient')

  # -------------------------------------
  # Horisontal gradient direction (Sobel)
  # -------------------------------------
  hy = np.array([[1, 0, -1],
                 [2, 0, -2],
                 [1, 0, -1]])
  hy_V = np.array([1, 2, 1])
  hy_H = np.array([1, 0, -1])

  gy_2d = conv2d(im, hy)
  gy_temp = conv1d_H(im, hy_H)
  gy_1d = conv1d_V(gy_temp, hy_V)
  gy_scipy = signal.convolve2d(im, hy, mode='same', boundary='fill')
  test_image_similarity(gy_1d, gy_2d, gy_scipy)

  fig_num = plot_image(gy_1d, fig_num, 'Horisontal gradient')

  cv2.imwrite(os.path.join(result_dir, 'grad_y_'+in_im_file), scale_uint8(gy_1d, 0, 255))

  # ------------------
  # Gradient magnitude
  # ------------------
  g_magn = np.sqrt(gx_1d*gx_1d + gy_1d*gy_1d)

  cv2.imwrite(os.path.join(result_dir, 'grad_magn_'+in_im_file), scale_uint8(g_magn, 0, 255))

  fig_num = plot_image(g_magn, fig_num, 'Gradient magnitude')

  # --------------
  # Gradient angle
  # --------------

  # This is rotated such that
  #
  #         180
  #          |
  #    270 ----- 90
  #          |
  #          0
  #
  g_angle = np.arctan2(gy_1d, gx_1d)

  cv2.imwrite(os.path.join(result_dir, 'grad_angle_'+in_im_file), scale_uint8(g_angle, 0, 255))

  fig_num = plot_image(g_angle, fig_num, 'Gradient angle')

  plt.show()

if __name__ == "__main__":
  main()
